#define com_maroontress_libcgi_db_Box_IMPORT
#define com_maroontress_libcgi_db_BoxKit_IMPORT
#define com_maroontress_libcgi_db_Tree_IMPORT

#include <sys/types.h>
#include <sys/param.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <com/maroontress/libcgi/db/Box.h>
#include <com/maroontress/libcgi/db/BoxKit.h>
#include <com/maroontress/libcgi/db/Tree.h>
#include "db.h"

#define HAS_BSD_OPEN (!defined(_POSIX_SOURCE)				\
                      && defined(O_SHLOCK) && defined(O_EXLOCK))
#if !HAS_BSD_OPEN
#include <sys/file.h>
#define O_EXLOCK 0
#define O_SHLOCK 0
#endif

struct Tree {
    DB *db;
    struct Box *box;
    int hasError;
};

struct Tree_Mode {
    int flags;
    int mode;
    int operation;
};

#ifdef WITH_BDB4
#define FLAGS_SH (O_RDONLY | O_SHLOCK)
#else
#define FLAGS_SH (O_CREAT | O_RDONLY | O_SHLOCK)
#endif
#define FLAGS_EX (O_CREAT | O_RDWR | O_EXLOCK)

static const struct Tree_Mode MODE_SHAERD = {FLAGS_SH, 0666, LOCK_SH};
const struct Tree_Mode *const Tree_MODE_SHARED = &MODE_SHAERD;

static const struct Tree_Mode MODE_EXCLUSIVE = {FLAGS_EX, 0666, LOCK_EX};
const struct Tree_Mode *const Tree_MODE_EXCLUSIVE = &MODE_EXCLUSIVE;

struct Tree *
Tree_create(const char *path, const struct Tree_Mode *m)
{
    struct Tree *t;

    if ((t = (struct Tree *)malloc(sizeof(*t))) == NULL)
	return NULL;
    t->db = NULL;
    t->box = NULL;
    t->hasError = 0;

    if ((t->db = dbopen(path, m->flags, m->mode, DB_BTREE, NULL)) == NULL) {
	Tree_free(t);
	return NULL;
    }
#if !HAS_BSD_OPEN
    flock(t->db->fd(t->db), m->operation);
#endif
    return t;
}

void
Tree_free(struct Tree *t)
{
    if (t == NULL)
	return;
    if (t->db != NULL)
	t->db->close(t->db);
    Box_free(t->box);
    free(t);
}

struct BoxKit *
Tree_getBoxKit(struct Tree *t, const char *key)
{
    struct BoxKit *kit;
    DB *db = t->db;
    DBT k, v;
    int s;

    k.data = (void *)key;
    k.size = strlen(key) + 1;
    if ((s = db->get(db, &k, &v, 0)) < 0
	|| (kit = BoxKit_create(t, key)) == NULL) {
	return NULL;
    }
    if (s == 0 && BoxKit_load(kit, v.data, v.size) < 0) {
	BoxKit_purge(kit);
	return NULL;
    }
    return kit;
}

int
Tree_delete(struct Tree *t, const char *key)
{
    DB *db = t->db;
    DBT k;
    int s;

    k.data = (void *)key;
    k.size = strlen(key) + 1;
    if ((s = db->del(db, &k, 0)) != 0)
	return s;
    return db->sync(db, 0);
}

static struct Box *
updateBox(struct Tree *t, const char *key, const void *data, size_t size)
{
    Box_free(t->box);
    if ((t->box = Box_create(key, data, size)) == NULL) {
	t->hasError = 1;
	return NULL;
    }
    return t->box;
}

struct Box *
Tree_getBox(struct Tree *t, const char *key)
{
    DB *db = t->db;
    DBT k, v;
    int s;

    k.data = (void *)key;
    k.size = strlen(key) + 1;
    if ((s = db->get(db, &k, &v, 0)) < 0) {
	t->hasError = 1;
	return NULL;
    }
    if (s == 1) {
	return NULL;
    }
    return updateBox(t, key, v.data, v.size);
}

static struct Box *
scanSequentialBox(struct Tree *t, int flags)
{
    DB *db = t->db;
    DBT k, v;
    int s;

    if ((s = db->seq(db, &k, &v, flags)) < 0) {
	t->hasError = 1;
	return NULL;
    }
    if (s == 1) {
	return NULL;
    }
    if (memchr(k.data, 0, k.size) == NULL) {
	t->hasError = 1;
	return NULL;
    }
    return updateBox(t, (char *)k.data, v.data, v.size);
}

struct Box *
Tree_scanFirstBox(struct Tree *t)
{
    return scanSequentialBox(t, R_FIRST);
}

struct Box *
Tree_scanNextBox(struct Tree *t)
{
    return scanSequentialBox(t, R_NEXT);
}

struct Box *
Tree_scanLastBox(struct Tree *t)
{
    return scanSequentialBox(t, R_LAST);
}

struct Box *
Tree_scanPreviousBox(struct Tree *t)
{
    return scanSequentialBox(t, R_PREV);
}

struct Box *
Tree_scanBox(struct Tree *t, const char *key)
{
    DB *db = t->db;
    DBT k, v;
    int s;

    k.data = (void *)key;
    k.size = strlen(key) + 1;
    if ((s = db->seq(db, &k, &v, R_CURSOR)) < 0) {
	t->hasError = 1;
	return NULL;
    }
    if (s == 1) {
	return NULL;
    }
    if (memchr(k.data, 0, k.size) == NULL) {
	t->hasError = 1;
	return NULL;
    }
    return updateBox(t, key, v.data, v.size);
}

int
Tree_put(struct Tree *t, const char *key, const void *data, size_t size)
{
    DB *db = t->db;
    DBT k, v;

    k.data = (void *)key;
    k.size = strlen(key) + 1;
    v.data = (void *)data;
    v.size = size;
    if (db->put(db, &k, &v, 0) < 0)
	return -1;
    return db->sync(db, 0);
}

int
Tree_hasError(struct Tree *t)
{
    return t->hasError;
}
